package eu.DECIDEh2020.mcsla.editor;

/*
* Copyright (c) 2017 Fraunhofer FOKUS.
* All rights reserved. This program and the accompanying materials
* are made available under the terms of the
* <<licensing_schema_to_be_decided>> which accompanies
* this distribution, and is available at
* https://opensource.org/licenses/EPL-2.0
*
* Contributors:
*
* Majid Salehi Ghamsari Fraunhofer FOKUS
**Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
*/
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.UUID;

import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.web.client.RestTemplate;

import eu.DECIDEh2020.mcsla.beans.Csp;
import eu.DECIDEh2020.mcsla.beans.Mcsla;
import eu.DECIDEh2020.mcsla.beans.Metric;
import eu.DECIDEh2020.mcsla.beans.MicroserviceSLA;
import eu.DECIDEh2020.mcsla.beans.MicroserviceSLO;
import eu.DECIDEh2020.mcsla.beans.MicroserviceSQO;
import eu.DECIDEh2020.mcsla.beans.ProjectConfig;
import eu.DECIDEh2020.mcsla.beans.Remedy;
import eu.DECIDEh2020.mcsla.beans.UnderlyingMetricRef;
import eu.DECIDEh2020.mcsla.beans.ViolationTriggerRule;
import eu.DECIDEh2020.mcsla.editor.configuration.AppConfig;
import eu.DECIDEh2020.mcsla.editor.service.ConfigFileService;

public class JerseyClient {
	// private static final String MCSLA_SRVICE_URL =
	// "http://localhost:8080/decide.mcsla.srv";
	// private static final String SRV_MCSLAS =
	// "http://localhost:8080/decide.mcsla.srv/mcslas";
	// private static final String SRV_MCSLA =
	// "http://localhost:8080/decide.mcsla.srv/mcsla";
	// private static final String SRV_CSPS =
	// "http://localhost:8080/decide.mcsla.srv/csps";
	// private static final String SRV_AGGREGATE =
	// "http://localhost:8080/decide.mcsla.srv/aggregate";
	// private static final String SRV_CONFIG =
	// "http://localhost:8080/decide.mcsla.srv/config";

	private String m_serviceURL = "http://localhost:8080/decide.mcsla.srv";

	public JerseyClient() {
		super();
		System.out.println("JerseyClient.......");
		ConfigFileService configService = getConfigService();
		this.m_serviceURL = configService.getServiceUrl();
	}

	private ConfigFileService getConfigService() {
		AbstractApplicationContext context = new AnnotationConfigApplicationContext(AppConfig.class);
		ConfigFileService configService = (ConfigFileService) context.getBean("configFileService");
		context.close();
		return configService;
	}

	@SuppressWarnings("unused")
	private static MicroserviceSLA addSQOs(MicroserviceSLA sla) {
		UUID uuid = UUID.randomUUID();

		List<MicroserviceSQO> sqos = new ArrayList<>();
		MicroserviceSQO sqo = new MicroserviceSQO();
		sqo.setMicroservice_SQO_id(uuid.toString());
		sqo.set__comment("CPU Uitlization will not be less then 85% of the time Over a Day");
		sqo.setTermName("CPU-utilization");
		sqo.setMonitoringResult("result");
		sqo.setValue("95");
		sqo.setUnit("percentage");
		sqo.setCalculationFormula("CPU-utilization LT 85");
		List<ViolationTriggerRule> vtrs = new ArrayList<>();
		ViolationTriggerRule vtr = new ViolationTriggerRule();
		vtr.setBreaches_count(2);
		vtr.setViointerval(30);
		vtrs.add(vtr);
		sqo.setViolationTriggerRule(vtrs);
		Remedy remedy = new Remedy();
		remedy.setType("discount");
		remedy.setUnit("%");
		remedy.setValidity("P1D");
		remedy.setValue(5);
		sqo.setRemedy(remedy);

		List<Metric> metrics = new ArrayList<>();
		Metric metric = new Metric();
		metric.setMetric_id("CSA_002");
		metric.setDescriptor("CloudServiceAvailability");
		UnderlyingMetricRef underlyingMetricRef = new UnderlyingMetricRef();
		underlyingMetricRef.setRefid("TQD_001");
		metric.setUnderlyingMetricRef(underlyingMetricRef);
		metrics.add(metric);
		sqo.setMetrics(metrics);
		sqos.add(sqo);
		sla.setMicroservice_SQOs(sqos);
		return sla;
	}

	public Mcsla createMCSLAexample(MicroserviceSLO slo) {

		UUID uid = UUID.randomUUID();
		String randomUUIDString = uid.toString();
		Mcsla mcsla = new Mcsla();
		mcsla.setmCSLA_id(uid);
		String timeStamp = new SimpleDateFormat("yyyy.MM.dd HH.mm").format(Calendar.getInstance().getTime());
		mcsla.setDescription(
				"This is MCSLA dummy description line created from createMCSLAexample method at:" + timeStamp);
		mcsla.setVisibility("public");
		mcsla.setValidityPeriod("this is Validity Period line.");
		List<MicroserviceSLA> slas = new ArrayList<>();

		UUID uuid = UUID.randomUUID();
		MicroserviceSLA sla = new MicroserviceSLA();
		sla.setMicroservice_SLA_id(uuid.toString());

		uuid = UUID.randomUUID();
		randomUUIDString = uuid.toString();
		// TODO fix this need valid CSP
		sla.setCsp_id(randomUUIDString);
		// TODO fix this need valid setValidityPeriod
		sla.setValidityPeriod(Integer.toString(5));
		sla.setVisibility("public");
		uuid = UUID.randomUUID();
		randomUUIDString = uuid.toString();
		sla.setMicroservice_id(randomUUIDString);

		if (slo != null) {
			sla.set__comment(slo.__comment);
			List<MicroserviceSLO> slos = new ArrayList<>();
			slos.add(slo);
			sla.setMicroservice_SLOs(slos);

		}
		slas.add(sla);
		mcsla.setMicroservice_SLAs(slas);
		return mcsla;

	}

	public Mcsla buildMCSLAwithslos(Mcsla mcsla, List<MicroserviceSLO> slos, List<MicroserviceSQO> sqos) {
		// TODO fix this default line
		mcsla.setValidityPeriod("this is Validity Period line.");
		List<MicroserviceSLA> slas = new ArrayList<>();

		UUID uuid = UUID.randomUUID();
		MicroserviceSLA sla = new MicroserviceSLA();
		sla.setMicroservice_SLA_id(uuid.toString());

		uuid = UUID.randomUUID();
		String randomUUIDString = uuid.toString();
		// TODO fix this need valid CSP
		sla.setCsp_id(randomUUIDString);
		// TODO fix this need valid setValidityPeriod
		sla.setValidityPeriod(Integer.toString(5));
		sla.setVisibility("public");
		uuid = UUID.randomUUID();
		randomUUIDString = uuid.toString();
		sla.setMicroservice_id(randomUUIDString);

		if (slos != null) {
			sla.setMicroservice_SLOs(slos);
		}
		if (sqos != null) {
			sla.setMicroservice_SQOs(sqos);
		}
		slas.add(sla);
		mcsla.setMicroservice_SLAs(slas);
		return mcsla;

	}

	public ProjectConfig getProjectConfig() {
		RestTemplate restTemplate = new RestTemplate();
		ProjectConfig pconfig = restTemplate.getForObject(m_serviceURL + "/config", ProjectConfig.class);
		System.out.println("JerseyClient:getAppConfig:" + pconfig.getGitusername() + ", " + pconfig.getGitremotepath());
		return pconfig;
	}

	public static void main(String[] args) {

		System.out.println("####################TEST MAIN START####################");
		JerseyClient jerseyClient = new JerseyClient();
		// jerseyClient.getMcslaDetails();
		// get currently used CSPs returns CSP SLAs
		MicroserviceSLA[] slas = jerseyClient.getCspSLAs();
		System.out.println("Main:got CSP SLAs:" + slas.length);
		// Prepare SLOs to send to MCSLA Aggregator
		MicroserviceSLO[] slos = jerseyClient.buildSLOs(slas);
		System.out.println("Main:got list SLOs:" + slos.length);
		// Return aggregated SLO
		MicroserviceSLO aggregatedSLO = jerseyClient.applyAggregationRules(slos);
		if (aggregatedSLO != null)
			System.out.println("Main:got aggregated SLO:" + aggregatedSLO.termName);
		else
			System.out.println("Main:got aggregated SLO is NULL :-(");
		// if (true)
		// return;

		Mcsla mcsla = jerseyClient.createMCSLAexample(aggregatedSLO);
		jerseyClient.addMcsla(mcsla);
		System.out.println("Main:addMcsla done.");

		DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
		Date date = new Date();
		mcsla.setDescription("this is MCSLA UPDATED at:" + dateFormat.format(date) + " description line.");
		mcsla = jerseyClient.updateMcsla(mcsla);
		if (mcsla == null) {
			System.out.println("Main:got updated mcsla is NULL :-(");
			return;
		} else
			System.out.println("Main:updateMcsla done.");

		mcsla = jerseyClient.getMcslaById(mcsla.mCSLA_id.toString());
		System.out.println("Main:getMcslaById:" + mcsla.getmCSLA_id() + ", " + mcsla.getDescription());
		// jerseyClient.deleteMcsla(mcsla.getmCSLA_id());
		System.out.println("####################TEST MAIN DONE####################");

	}

	public void addMcsla(Mcsla mcsla) {
		// It is used to create new resource. Method POST
		RestTemplate restTemplate = new RestTemplate();
		HttpEntity<Mcsla> request = new HttpEntity<>(mcsla);
		Mcsla resMcsla = restTemplate.postForObject(m_serviceURL + "/mcslas", request, Mcsla.class);
		if (resMcsla == null)
			System.out.println("JerseyClient:addMcsla:FAILED");
		else
			System.out.println("JerseyClient:addMcsla:" + resMcsla.getmCSLA_id() + ", " + resMcsla.getDescription());
	}

	public MicroserviceSLO applyAggregationRules(MicroserviceSLO[] slos) {
		System.out.printf("JerseyClient:applyAggregationRules to %s SLOs %n", slos.length);
		RestTemplate restTemplate = new RestTemplate();
		HttpEntity<MicroserviceSLO[]> requestEntity = new HttpEntity<MicroserviceSLO[]>(slos);
		HttpEntity<MicroserviceSLO> response = restTemplate.exchange(m_serviceURL + "/aggregate", HttpMethod.POST,
				requestEntity, MicroserviceSLO.class);
		MicroserviceSLO resSLO = response.getBody();
		System.out.println(
				"JerseyClient:applyAggregationRules:" + resSLO.getMicroservice_SLO_id() + ", " + resSLO.getTermName());
		return resSLO;
	}

	public MicroserviceSLO[] buildSLOs(MicroserviceSLA[] slas) {
		List<MicroserviceSLO> sloslist = new ArrayList<>();
		for (MicroserviceSLA sla : slas) {
			if (sla.getMicroservice_SLOs() != null)
				sloslist.addAll(sla.getMicroservice_SLOs());
		}
		MicroserviceSLO[] slos = new MicroserviceSLO[sloslist.size()];
		slos = sloslist.toArray(slos);
		return slos;
	}

	public MicroserviceSQO[] buildSQOs(MicroserviceSLA[] slas) {
		List<MicroserviceSQO> sqoslist = new ArrayList<>();
		for (MicroserviceSLA sla : slas) {
			if (sla.getMicroservice_SQOs() != null)
				sqoslist.addAll(sla.getMicroservice_SQOs());
		}
		MicroserviceSQO[] sqos = new MicroserviceSQO[sqoslist.size()];
		sqos = sqoslist.toArray(sqos);
		return sqos;
	}

	public void deleteMcsla(UUID uid) {
		RestTemplate restTemplate = new RestTemplate();
		String entityUrl = m_serviceURL + "/mcsla" + uid.toString();
		restTemplate.delete(entityUrl);
	}

	public MicroserviceSLA[] getCspSLAs() {
		RestTemplate restTemplate = new RestTemplate();
		ResponseEntity<Csp[]> responseEntity = restTemplate.getForEntity(m_serviceURL + "/csps", Csp[].class);
		Csp[] csplist = responseEntity.getBody();

		List<MicroserviceSLA> slaslist = new ArrayList<>();
		// for each CSP get SLAs
		for (int index = 0; index < csplist.length; index++) {
			ResponseEntity<MicroserviceSLA[]> responseEntitylist = restTemplate.getForEntity(
					m_serviceURL + "/csps" + "/{id}", MicroserviceSLA[].class, csplist[index].getmCSP_id());
			MicroserviceSLA[] slas = responseEntitylist.getBody();
			List<MicroserviceSLA> cspslas = new ArrayList<MicroserviceSLA>(Arrays.asList(slas));
			slaslist.addAll(cspslas);
		}

		System.out.println("JerseyClient:getCspSLAs:" + slaslist.size());
		return slaslist.toArray(new MicroserviceSLA[slaslist.size()]);
	}

	public Mcsla getMcslaById(String id) {
		RestTemplate restTemplate = new RestTemplate();
		Mcsla mcsla = restTemplate.getForObject(m_serviceURL + "mcsla" + "/{id}", Mcsla.class, id);
		System.out.println("JerseyClient:getMcslaById:" + mcsla.getmCSLA_id() + ", " + mcsla.getDescription());
		return mcsla;
	}

	public Mcsla getMcsla() {
		RestTemplate restTemplate = new RestTemplate();
		Mcsla mcsla = restTemplate.getForObject(m_serviceURL + "/mcsla", Mcsla.class);
		System.out.println("JerseyClient:getMcsla:" + mcsla.getmCSLA_id() + ", " + mcsla.getDescription());
		return mcsla;
	}

	public void getMcslaDetails() {
		RestTemplate restTemplate = new RestTemplate();
		ResponseEntity<Mcsla[]> responseEntity = restTemplate.getForEntity(m_serviceURL + "/mcslas", Mcsla[].class);
		Mcsla[] list = responseEntity.getBody();
		// HttpStatus statusCode = responseEntity.getStatusCode();
		// System.out.println("HttpStatus:getMcslaDetails:" +
		// statusCode.value());

		for (int index = 0; index < list.length; index++) {
			System.out.println(list[index].getmCSLA_id() + ", " + list[index].getDescription());
		}
	}

	public Mcsla updateMcsla(Mcsla mcsla) {
		// It is generally used to update resource, METHOD PUT
		System.out.println("JerseyClient:updateMcsla:" + mcsla.getmCSLA_id() + ", " + mcsla.getDescription());
		RestTemplate restTemplate = new RestTemplate();
		HttpEntity<Mcsla> requestEntity = new HttpEntity<Mcsla>(mcsla);
		HttpEntity<Mcsla> response = restTemplate.exchange(m_serviceURL + "/mcslas", HttpMethod.PUT, requestEntity,
				Mcsla.class);
		Mcsla resMcsla = response.getBody();
		if (resMcsla == null)
			System.out.println("JerseyClient:updateMcsla:FAILED");
		return resMcsla;
	}
}
