package eu.DECIDEh2020.mcsla.service.service;
/*
* Copyright (c) 2017 Fraunhofer FOKUS.
* All rights reserved. This program and the accompanying materials
* are made available under the terms of the
* <<licensing_schema_to_be_decided>> which accompanies
* this distribution, and is available at
* https://opensource.org/licenses/EPL-2.0
*
* Contributors:
*
* Majid Salehi Ghamsari Fraunhofer FOKUS
**Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
*/
import java.io.IOException;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.UUID;

import org.eclipse.jgit.api.errors.GitAPIException;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;

import eu.DECIDEh2020.mcsla.beans.Csp;
import eu.DECIDEh2020.mcsla.beans.Mcsla;
import eu.DECIDEh2020.mcsla.beans.Metric;
import eu.DECIDEh2020.mcsla.beans.MicroserviceSLA;
import eu.DECIDEh2020.mcsla.beans.MicroserviceSLO;
import eu.DECIDEh2020.mcsla.beans.MicroserviceSQO;
import eu.DECIDEh2020.mcsla.beans.Remedy;
import eu.DECIDEh2020.mcsla.beans.UnderlyingMetricRef;
import eu.DECIDEh2020.mcsla.beans.ViolationTriggerRule;
import eu.DECIDEh2020.mcsla.service.configuration.AppConfig;
import eu.DECIDEh2020.mcsla.service.configuration.ProjectConfig;
import eu.DECIDEh2020.mcsla.service.git.GitControl;

/*
 * It is just a helper class which should be replaced by database implementation.
 * It is not very well written class, it is just used for demonstration.
 */

public class McslaService {

	private static final String SERVICE_COMPOSITION_DIFFEREN_VENDORS_PATTERN = "ServiceCompositionDifferenVendorsPattern";
	private static final String SERVICE_COMPOSITION_REPLICATION_PATTERN = "ServiceCompositionReplicationPattern";
	private static final String SERVICE_COMPOSITION__SIMPLE_PATTERN = "ServiceCompositionPattern";
	static HashMap<UUID, Mcsla> mcslaIdMap = getMcslaIdMap();
	static HashMap<UUID, Csp> cspIdMap = getCspIdMap();
	// private static Logger logger = LoggerFactory.getLogger(McslaService.class);

	private static MicroserviceSLA addSLOs(MicroserviceSLA sla) {

		List<MicroserviceSLO> slos = new ArrayList<>();
		MicroserviceSLO slo = createExampleUPtimeSLO("99,9");
		slo.setCsp_id(sla.getCsp_id());
		slo.setCsp_name(sla.getCsp_name());
		slos.add(slo);
		slo = createExampleUtilizationSLO();
		slo.setCsp_id(sla.getCsp_id());
		slo.setCsp_name(sla.getCsp_name());
		slos.add(slo);
		sla.setMicroservice_SLOs(slos);
		return sla;
	}

	private static Csp addSomeSLAs2CSP(Csp csp, int j) {

		UUID uid = UUID.randomUUID();
		String randomUUIDString = uid.toString();

		List<MicroserviceSLA> slas = new ArrayList<>();
		int i = 0;
		do {
			i++;
			UUID uuid = UUID.randomUUID();
			MicroserviceSLA sla = new MicroserviceSLA();
			sla.setMicroservice_SLA_id(uuid.toString());
			sla.set__comment("UPtime from CSP " + csp.getName()
					+ " gives a clear picture of the overall performance of the cloud(UPTIME).");
			sla.setCsp_id(csp.getmCSP_id().toString());
			sla.setCsp_name(csp.getName());
			sla.setValidityPeriod(Integer.toString(i));
			sla.setVisibility("private");
			uuid = UUID.randomUUID();
			randomUUIDString = uuid.toString();
			sla.setMicroservice_id(randomUUIDString);
			sla = addSQOs(sla);
			sla = addSLOs(sla);
			slas.add(sla);
		} while (i < j);

		csp.setSlas(slas);

		return csp;

	}

	private static MicroserviceSLA addSQOs(MicroserviceSLA sla) {
		List<MicroserviceSQO> sqos = new ArrayList<>();
		MicroserviceSQO sqo = createExampleSQO("95");
		sqos.add(sqo);
		sqo = createExampleSQO("97");
		sqos.add(sqo);
		sla.setMicroservice_SQOs(sqos);
		return sla;
	}

	private static MicroserviceSLO createExampleUPtimeSLO(String uptime) {
		MicroserviceSLO slo = new MicroserviceSLO();
		UUID uuid = UUID.randomUUID();
		slo.setMicroservice_SLO_id(uuid.toString());
		slo.set__comment("UPtime will not be less then " + uptime + "% of the time Over a Day");
		slo.setTermName("Uptime-SLO");
		slo.setMonitoringResult("result");
		slo.setValue(uptime);
		slo.setUnit("percentage");
		slo.setCalculationFormula("Uptime GT " + uptime);
		List<ViolationTriggerRule> vtrs = new ArrayList<>();
		ViolationTriggerRule vtr = new ViolationTriggerRule();
		vtr.setBreaches_count(2);
		vtr.setViointerval(30);
		vtrs.add(vtr);
		slo.setViolationTriggerRule(vtrs);
		Remedy remedy = new Remedy();
		remedy.setType("discount");
		remedy.setUnit("%");
		remedy.setValidity("P1D");
		remedy.setValue(5);
		slo.setRemedy(remedy);

		List<Metric> metrics = new ArrayList<>();
		Metric metric = new Metric();
		metric.setMetric_id("CSA_002");
		metric.setDescriptor("CloudServiceAvailability");
		UnderlyingMetricRef underlyingMetricRef = new UnderlyingMetricRef();
		underlyingMetricRef.setRefid("TQD_001");
		metric.setUnderlyingMetricRef(underlyingMetricRef);
		metrics.add(metric);
		slo.setMetrics(metrics);
		return slo;
	}

	private static MicroserviceSLO createExampleUtilizationSLO() {
		MicroserviceSLO slo = new MicroserviceSLO();
		UUID uuid = UUID.randomUUID();
		slo.setMicroservice_SLO_id(uuid.toString());
		slo.set__comment("CPU Utilization will not be less then 95% of the time Over a Day");
		slo.setTermName("CPU-utilization-SLO");
		slo.setMonitoringResult("result");
		slo.setValue("95");
		slo.setUnit("percentage");
		slo.setCalculationFormula("CPU-utilization LT 95");
		List<ViolationTriggerRule> vtrs = new ArrayList<>();
		ViolationTriggerRule vtr = new ViolationTriggerRule();
		vtr.setBreaches_count(2);
		vtr.setViointerval(30);
		vtrs.add(vtr);
		slo.setViolationTriggerRule(vtrs);
		Remedy remedy = new Remedy();
		remedy.setType("discount");
		remedy.setUnit("%");
		remedy.setValidity("P1D");
		remedy.setValue(5);
		slo.setRemedy(remedy);

		List<Metric> metrics = new ArrayList<>();
		Metric metric = new Metric();
		metric.setMetric_id("CSA_002");
		metric.setDescriptor("CloudServiceAvailability");
		UnderlyingMetricRef underlyingMetricRef = new UnderlyingMetricRef();
		underlyingMetricRef.setRefid("TQD_001");
		metric.setUnderlyingMetricRef(underlyingMetricRef);
		metrics.add(metric);
		slo.setMetrics(metrics);
		return slo;
	}

	private static MicroserviceSQO createExampleSQO(String value) {
		MicroserviceSQO sqo = new MicroserviceSQO();
		UUID uuid = UUID.randomUUID();
		sqo.setMicroservice_SQO_id(uuid.toString());
		sqo.set__comment("CPU Uitlization will not be less then 95% of the time Over a Day");
		sqo.setTermName("CPU-utilization-SQO");
		sqo.setMonitoringResult("result");
		sqo.setValue(value);
		sqo.setUnit("percentage");
		sqo.setCalculationFormula("CPU-utilization LT 95");
		List<ViolationTriggerRule> vtrs = new ArrayList<>();
		ViolationTriggerRule vtr = new ViolationTriggerRule();
		vtr.setBreaches_count(2);
		vtr.setViointerval(30);
		vtrs.add(vtr);
		sqo.setViolationTriggerRule(vtrs);
		Remedy remedy = new Remedy();
		remedy.setType("discount");
		remedy.setUnit("%");
		remedy.setValidity("P1D");
		remedy.setValue(5);
		sqo.setRemedy(remedy);

		List<Metric> metrics = new ArrayList<>();
		Metric metric = new Metric();
		metric.setMetric_id("CSA_002");
		metric.setDescriptor("CloudServiceAvailability");
		UnderlyingMetricRef underlyingMetricRef = new UnderlyingMetricRef();
		underlyingMetricRef.setRefid("TQD_001");
		metric.setUnderlyingMetricRef(underlyingMetricRef);
		metrics.add(metric);
		sqo.setMetrics(metrics);
		return sqo;
	}

	public static HashMap<UUID, Csp> getCspIdMap() {
		return cspIdMap;
	}

	public static HashMap<UUID, Mcsla> getMcslaIdMap() {
		return mcslaIdMap;
	}

	public McslaService() {
		super();

		System.out.println("McslaService start....");

		if (mcslaIdMap == null) {
			createMcslaHashmap();
		}
		if (cspIdMap == null) {
			createCspHashmap();
		}
	}

	public Mcsla addMcsla(Mcsla mcsla) {
		if (mcsla.getmCSLA_id() == null)
			mcsla.setmCSLA_id(UUID.randomUUID());
		mcslaIdMap.put(mcsla.getmCSLA_id(), mcsla);

		// Push to repository
		try {
			GitControl gc = new GitControl();
			gc.pushMcslaToRepo(mcsla);
		} catch (IOException | GitAPIException e) {
			System.out.println("McslaService:addMcsla:FAILED");
			e.printStackTrace();
			return null;
		}

		return mcsla;
	}

	public MicroserviceSLO aggregate(List<MicroserviceSLO> slos) {
		return doAggregation(slos, SERVICE_COMPOSITION__SIMPLE_PATTERN);
	}

	private MicroserviceSLO doAggregation(List<MicroserviceSLO> slos, String pattern) {
		System.out.println("McslaService:doAggregation:" + pattern);
		String operation = "SUMtype";
		double calculatedUptime = 100;
		if (pattern.equalsIgnoreCase(SERVICE_COMPOSITION__SIMPLE_PATTERN))
			operation = "SUMtype";
		if (pattern.equalsIgnoreCase(SERVICE_COMPOSITION_REPLICATION_PATTERN))
			operation = "MINtype";
		if (pattern.equalsIgnoreCase(SERVICE_COMPOSITION_DIFFEREN_VENDORS_PATTERN))
			operation = "ORtype";

		for (Iterator<MicroserviceSLO> iterator = slos.iterator(); iterator.hasNext();) {
			MicroserviceSLO microserviceSLO = (MicroserviceSLO) iterator.next();
			// TODO FIX THIS Uptime-SLO?
			if (microserviceSLO.termName.equalsIgnoreCase("Uptime-SLO"))
				calculatedUptime = caluculate(microserviceSLO.getValue(), operation, calculatedUptime);
		}
		MicroserviceSLO aggSlo = createExampleUPtimeSLO(String.valueOf(calculatedUptime));
		aggSlo.csp_name = "!Aggregated!";
		return aggSlo;
	}

	private static double round(double value, int precision) {
		int scale = (int) Math.pow(10, precision);
		return (double) Math.round(value * scale) / scale;
	}

	private double caluculate(String value, String operation, double olduptime) {
		// System.out.println("McslaService:caluculate:value:" + value + " oldtime:" +
		// olduptime);
		double downtime = 0;
		if (operation.equalsIgnoreCase("SUMtype")) {
			NumberFormat format = NumberFormat.getInstance(Locale.FRANCE);
			Number number = null;
			try {
				number = format.parse(value);
			} catch (ParseException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			double d = number.doubleValue();
			downtime = 100 - d;
		}
		// System.out.println("McslaService:caluculate:downtime:" + downtime);
		double newtime = olduptime - downtime;
		// System.out.println("McslaService:caluculate:newtime:" + newtime);

		return round(newtime, 1);
	}

	private void createCspHashmap() {
		cspIdMap = new HashMap<UUID, Csp>();
		// Creating some objects of CSP while initializing
		Csp csp1 = new Csp(UUID.randomUUID(), "Amazon");
		addSomeSLAs2CSP(csp1, 1);
		Csp csp2 = new Csp(UUID.randomUUID(), "Azure");
		addSomeSLAs2CSP(csp2, 1);
		Csp csp3 = new Csp(UUID.randomUUID(), "IBM");
		addSomeSLAs2CSP(csp3, 1);
		// Csp csp4 = new Csp(UUID.randomUUID(), "Telekom");
		// addSomeSLAs2CSP(csp4, 1);

		cspIdMap.put(csp1.getmCSP_id(), csp1);
		cspIdMap.put(csp2.getmCSP_id(), csp2);
		cspIdMap.put(csp3.getmCSP_id(), csp3);
		// cspIdMap.put(csp4.getmCSP_id(), csp4);

	}

	private void createMcslaHashmap() {

		mcslaIdMap = new HashMap<UUID, Mcsla>();
		// Creating some objects of Mcsla while initializing
		Mcsla m1 = new Mcsla(UUID.randomUUID(), "this is MCSLA1 description line.", "public",
				"this is MCSLA1 setValidityPeriod line.");
		Mcsla m2 = new Mcsla(UUID.randomUUID(), "this is MCSLA2 description line.", "public",
				"this is MCSLA2 setValidityPeriod line.");
		Mcsla m3 = new Mcsla(UUID.randomUUID(), "this is MCSLA3 description line.", "private",
				"this is MCSLA3 setValidityPeriod line.");
		Mcsla m4 = new Mcsla(UUID.randomUUID(), "this is MCSLA4 description line.", "public",
				"this is MCSLA4 setValidityPeriod line.");
		Mcsla m5 = new Mcsla(UUID.randomUUID(), "this is MCSLA5 description line.", "public",
				"this is MCSLA5 setValidityPeriod line.");

		mcslaIdMap.put(m1.getmCSLA_id(), m1);
		mcslaIdMap.put(m2.getmCSLA_id(), m2);
		mcslaIdMap.put(m3.getmCSLA_id(), m3);
		mcslaIdMap.put(m4.getmCSLA_id(), m4);
		mcslaIdMap.put(m5.getmCSLA_id(), m5);
	}

	public void deleteMcsla(String id) {
		UUID uid = UUID.fromString(id);
		mcslaIdMap.remove(uid);
	}

	public List<Csp> getAllCsps() {
		List<Csp> csps = new ArrayList<Csp>(cspIdMap.values());
		return csps;
	}

	public List<Mcsla> getAllMcslas() {
		ConfigFileService configService = getConfigService();
		//configService.readValues();
		System.out.println("McslaService:updateMcsla:id:" + configService.getGitremotepath());

		List<Mcsla> mcslas = new ArrayList<Mcsla>(mcslaIdMap.values());
		return mcslas;
	}

	private ConfigFileService getConfigService() {
		AbstractApplicationContext context = new AnnotationConfigApplicationContext(AppConfig.class);
		ConfigFileService configService = (ConfigFileService) context.getBean("configFileService");
		context.close();
		return configService;
	}

	public List<MicroserviceSLA> getCspSLAs(String id) {
		UUID uid = UUID.fromString(id);
		Csp csp = cspIdMap.get(uid);
		return csp.getSlas();
	}

	public Mcsla getMcsla(String id) {
		UUID uid = UUID.fromString(id);
		Mcsla mcsla = mcslaIdMap.get(uid);
		return mcsla;
	}

	public Mcsla updateMcsla(Mcsla mcsla) {
		UUID uid = mcsla.getmCSLA_id();
		if (uid == null)
			return null;
		System.out.println("McslaService:updateMcsla:id:" + uid);
		mcslaIdMap.put(uid, mcsla);

		// Push to repository
		try {
			GitControl gc = new GitControl();
			gc.pushMcslaToRepo(mcsla);
		} catch (IOException | GitAPIException e) {
			// TODO Auto-generated catch block
			System.out.println("McslaService:updateMcsla:FAILED");
			e.printStackTrace();
			return null;
		}
		System.out.println("McslaService:updateMcsla:--Done--");
		return mcsla;

	}

	public ProjectConfig getProjectConfig() {

		ConfigFileService configService = getConfigService();
		ProjectConfig projectConfig = new ProjectConfig(configService.getGitremotepath(),
				configService.getGitusername());
		return projectConfig;
	}

	// operation on GET MCSLAfrom repo
	public Mcsla getMcsla() {
		Mcsla mcsla;
		// Get from repository
		try {
			GitControl gc = new GitControl();
			mcsla = gc.getMcslaFromRepo();
		} catch (IOException | GitAPIException e) {
			// TODO Auto-generated catch block
			System.out.println("McslaService:getMcsla:FAILED");
			e.printStackTrace();
			return null;
		}
		System.out.println("McslaService:getMcsla:--Done--");
		return mcsla;
	}

}
